/*++

Module Name:

    ioctl.c


X36 IO handler.

Overview:

	This module is responsible for handling all the device control calls that
pass through this filter.  It performs the following functions:
	-> Tweaking the USB descriptors as they come up through the stack to add
	   the parameters necessary to create a virtual keyboard.
	-> Capturing an I/O request to reply with keyboard data on my own timers
	-> Replying with keyboard data
	-> Filter X36 joystick data

--*/

// #define DBG 1

#include <WDM.H>
#include "usbioctl.h"
#include "usbdi.h"
#include "usbdlib.h"
#include "filter.h"
#include "fakehid.h"




#ifdef ALLOC_PRAGMA
        #pragma alloc_text(PAGE, VA_Ioctl)
#endif


NTSTATUS VA_Ioctl(struct DEVICE_EXTENSION *  devExt, PIRP Irp)
{
    NTSTATUS            ntStatus = STATUS_SUCCESS;    
    PIO_STACK_LOCATION  IrpStack = IoGetCurrentIrpStackLocation(Irp);    
    ULONG dwControlCode = IrpStack->Parameters.DeviceIoControl.IoControlCode;
	PURB pUrb = (PURB)IrpStack->Parameters.Others.Argument1;

			    
    if (IOCTL_INTERNAL_USB_SUBMIT_URB == dwControlCode)
	{

		switch (pUrb->UrbHeader.Function)
		{
		    case URB_FUNCTION_BULK_OR_INTERRUPT_TRANSFER:
				if ((devExt->ReadIrp == NULL)&&(KeGetCurrentIrql()==DISPATCH_LEVEL)) // Don't capture if we're not really running.
				{
					DbgPrn(("Capturing keyboard read IRP.\n"));
					
					KeAcquireSpinLockAtDpcLevel(&devExt->ReadLock);
					devExt->ReadIrp = Irp;
					KeReleaseSpinLockFromDpcLevel(&devExt->ReadLock);
					
		            IncrementPendingActionCount(devExt);
		            
		            IoMarkIrpPending(Irp);
		          	ntStatus = STATUS_PENDING;
		          	
		          	if (devExt->NewData)
		          	{
		          		DbgPrn(("Filling request right away.  New data: %d\n",devExt->NewData));
		          		KeInsertQueueDpc(&devExt->ReturnDPC,NULL,NULL);
		          	}
	            }
	            else
	            {
	            	IoCopyCurrentIrpStackLocationToNext(Irp);
		    		IoSetCompletionRoutine(Irp, &ReportCompletion, NULL, TRUE, TRUE, TRUE);    
		    		ntStatus = IoCallDriver(devExt->topDevObj, Irp);
		    	}
				break;
		
		    case URB_FUNCTION_RESET_PIPE:
		    	DbgPrn(("Reset pipe.\n"));
				if ( devExt->ReadIrp != NULL )
				{
					KeCancelTimer(&devExt->Timer);
					CompleteRead( NULL, devExt, NULL, NULL);				
				}
				break ;
				
			case URB_FUNCTION_ABORT_PIPE:
				DbgPrn(("Abort pipe.\n"));
				KeCancelTimer(&devExt->Timer);
				KeAcquireSpinLockAtDpcLevel(&devExt->ReadLock);
				if (devExt->ReadIrp != NULL)				
				{
			   		devExt->ReadIrp->IoStatus.Status = STATUS_CANCELLED;
			    	IoCompleteRequest(devExt->ReadIrp, IO_NO_INCREMENT);	
					devExt->ReadIrp = NULL;
					DecrementPendingActionCount( devExt );
				}
				KeReleaseSpinLockFromDpcLevel(&devExt->ReadLock);				
				break;
		
		    case URB_FUNCTION_GET_DESCRIPTOR_FROM_DEVICE:
				IoCopyCurrentIrpStackLocationToNext(Irp) ;
				IoSetCompletionRoutine(Irp,(PIO_COMPLETION_ROUTINE)ConfigDescriptorCompletion, NULL, TRUE, TRUE,TRUE) ;
				ntStatus = IoCallDriver(devExt->topDevObj, Irp) ;
				break;
		
		     case URB_FUNCTION_GET_DESCRIPTOR_FROM_INTERFACE:
				IoCopyCurrentIrpStackLocationToNext(Irp) ;
				IoSetCompletionRoutine(Irp,(PIO_COMPLETION_ROUTINE)ReportDescriptorCompletion, NULL, TRUE,TRUE, TRUE) ;
				ntStatus = IoCallDriver(devExt->topDevObj, Irp) ;
				break;
				
			default:
				IoSkipCurrentIrpStackLocation(Irp);
				ntStatus = IoCallDriver(devExt->topDevObj, Irp);
			}
			
	}
	else
	{
		IoSkipCurrentIrpStackLocation(Irp);
		ntStatus = IoCallDriver(devExt->topDevObj, Irp);
	}

    return ntStatus;    
}


NTSTATUS ConfigDescriptorCompletion( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context )
{
	/* Should print the data in the system. */
    NTSTATUS            ntStatus = STATUS_SUCCESS;
    struct DEVICE_EXTENSION *  DeviceExtension = DeviceObject->DeviceExtension;
   	PURB pUrb = (PURB)IoGetCurrentIrpStackLocation(Irp)->Parameters.Others.Argument1;
   	struct _URB_CONTROL_TRANSFER   *pControlTransfer = (struct _URB_CONTROL_TRANSFER *) pUrb;
   	PUCHAR BufPtr=NULL;
   	PUCHAR Temp;
    unsigned long b,i,pos, tmp;
    
    
    if (Irp->PendingReturned) {
        IoMarkIrpPending( Irp );
    }
    

	// OK,we need to (a) snapture the report id; then (b) snapture the interrupts themselves.
	// To get the report ID we need to know which one to watch for.
	// So, first we have to capture the HID descriptor and parse that.  It comes between the configuration and interface descriptors.


	if (pUrb != NULL)
	{
			// Figure out what packet this was.
			
		DbgPrn(("  SetupPacket          :"));
		for(b=0; b<sizeof(pControlTransfer->SetupPacket); b++)
			DbgPrn((" %02x", pControlTransfer->SetupPacket[b]));
		DbgPrn(("\n"));

		BufPtr=NULL;
		if ((PUCHAR)pControlTransfer->TransferBuffer)
		{
			BufPtr=(PUCHAR)pControlTransfer->TransferBuffer;
		}
		else if(pControlTransfer->TransferBufferMDL)
		{
			BufPtr = (PUCHAR)MmGetSystemAddressForMdl(pControlTransfer->TransferBufferMDL);
		}
		if (BufPtr==NULL)
		{	// Confused.  Abort.
			DbgPrn(("Can't find buffer, aborting.\n"));
			return STATUS_SUCCESS;
		}
								
		if ((pControlTransfer->SetupPacket[3]==0x02)&&((pControlTransfer->SetupPacket[6]>0x09)||(pControlTransfer->SetupPacket[7]>0)))
		{
			DbgPrn(("Complete descriptor transferred.  Locating HID descriptor.\n"));
			
			for(pos=0;((BufPtr[pos+1]!=0x21)&&(pos<pControlTransfer->TransferBufferLength));pos=pos+BufPtr[pos]);
			
			if (pos>=pControlTransfer->TransferBufferLength)
			{
				DbgPrn(("Can't find HID descriptor, aborting.\n"));
				return STATUS_SUCCESS;
			}
			else
			{
				DbgPrn(("HID descriptor located at position %x.\n",pos));
			}
				
			DeviceExtension->ReportDescType = BufPtr[pos+6];
			DeviceExtension->ReportBufSize = BufPtr[pos+7];
			BufPtr[pos+7]+=sizeof(MyReportDescriptor)+2;
										
										
			
			if (Temp = (PUCHAR)MmGetSystemAddressForMdl(pControlTransfer->TransferBufferMDL))
				RtlCopyMemory(Temp,BufPtr,pControlTransfer->TransferBufferLength);
		}   
	}
	return STATUS_SUCCESS;
}



NTSTATUS ReportDescriptorCompletion( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context )
{
	/* Should print the data in the system. */
    NTSTATUS            ntStatus = STATUS_SUCCESS;
    struct DEVICE_EXTENSION *  DeviceExtension = DeviceObject->DeviceExtension;
   	PURB pUrb = (PURB)IoGetCurrentIrpStackLocation(Irp)->Parameters.Others.Argument1;
   	struct _URB_CONTROL_TRANSFER   *pControlTransfer = (struct _URB_CONTROL_TRANSFER *) pUrb;
   	PUCHAR BufPtr=NULL;
   	PUCHAR Temp;
    unsigned long b,i,pos, tmp;
    
    
    if (Irp->PendingReturned) {
        IoMarkIrpPending( Irp );
    }
    

	// OK,we need to (a) snapture the report id; then (b) snapture the interrupts themselves.
	// To get the report ID we need to know which one to watch for.
	// So, first we have to capture the HID descriptor and parse that.  It comes between the configuration and interface descriptors.


	if (pUrb != NULL)
	{
			// Figure out what packet this was.
			
		DbgPrn(("  SetupPacket          :"));
		for(b=0; b<sizeof(pControlTransfer->SetupPacket); b++)
			DbgPrn((" %02x", pControlTransfer->SetupPacket[b]));
		DbgPrn(("\n"));

		BufPtr=NULL;
		if ((PUCHAR)pControlTransfer->TransferBuffer)
		{
			BufPtr=(PUCHAR)pControlTransfer->TransferBuffer;
		}
		else if(pControlTransfer->TransferBufferMDL)
		{
			BufPtr = (PUCHAR)MmGetSystemAddressForMdl(pControlTransfer->TransferBufferMDL);
		}
		if (BufPtr==NULL)
		{	// Confused.  Abort.
			DbgPrn(("Can't find buffer, aborting.\n"));
			return STATUS_SUCCESS;
		}
								


		if (pControlTransfer->SetupPacket[3]==DeviceExtension->ReportDescType)
		{
			DbgPrn(("Report Descriptor.\n"));

			
			// Fix axis maxima/minima.
			pos=9;
			for(i=0; i<6; i++)
			{
				if (DeviceExtension->Configurations[0][0][0]->Axis[i].Min)
					BufPtr[pos]=(unsigned char)DeviceExtension->Configurations[0][0][0]->Axis[i].Min;
				pos+=(i<2)?3:2;
				if (DeviceExtension->Configurations[0][0][0]->Axis[i].Max)
					*(unsigned short*)(&BufPtr[pos])=(unsigned short)DeviceExtension->Configurations[0][0][0]->Axis[i].Max;
				pos+=10;
			}
			
			// Fix throttle type.
			BufPtr[35]=0x32;
			
			
			RtlMoveMemory(BufPtr+8,BufPtr+6,pControlTransfer->TransferBufferLength-6);
			
			BufPtr[6]=0x85;			BufPtr[7]=0x01;	
			pControlTransfer->TransferBufferLength+=2;
			RtlCopyMemory(&BufPtr[pControlTransfer->TransferBufferLength],MyReportDescriptor, sizeof(MyReportDescriptor));
			pControlTransfer->TransferBufferLength+=sizeof(MyReportDescriptor);
			
		}
	}
	return STATUS_SUCCESS;
}



NTSTATUS ReportCompletion( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context )
{
	/* Should print the data in the system. */
    NTSTATUS            ntStatus = STATUS_SUCCESS;
    struct DEVICE_EXTENSION *  DeviceExtension = DeviceObject->DeviceExtension;
   	PURB pUrb = (PURB)IoGetCurrentIrpStackLocation(Irp)->Parameters.Others.Argument1;
	struct _URB_BULK_OR_INTERRUPT_TRANSFER *pBulkOrInterruptTransfer = (struct _URB_BULK_OR_INTERRUPT_TRANSFER *) pUrb;
   	PUCHAR BufPtr=NULL;
   	PUCHAR Temp;
   	LARGE_INTEGER LocalTime;
   	QENTRY Current;
   	unsigned char Delta[0x0b];
   	unsigned char switch1,switch2,shift;
    unsigned long b,i,pos, tmp;
    
    
    if (Irp->PendingReturned) {
        IoMarkIrpPending( Irp );
    }
    
    
    // Lag avoidance routine.
	KeQuerySystemTime(&LocalTime);
	

	// OK,we need to (a) snapture the report id; then (b) snapture the interrupts themselves.
	// To get the report ID we need to know which one to watch for.
	// So, first we have to capture the HID descriptor and parse that.  It comes between the configuration and interface descriptors.
	DbgPrn(("Joystick sending data packet.\n"));

	if ((pUrb != NULL)&&(pBulkOrInterruptTransfer->TransferBufferLength == 0x0b))
	{
		// This is the stick reporting its packet.
		// DbgPrn(("Stick reporting packet.\n"));
		
		
						
		
		// Do nothing at this point.  Later we'll mask off buttons.  Need to parse report descriptor first.
		// Force mask off buttons immediately after the ones lit in the packets.
		
		BufPtr = NULL;
	
		if ((PUCHAR)pBulkOrInterruptTransfer->TransferBuffer)
		{
			BufPtr=(PUCHAR)pBulkOrInterruptTransfer->TransferBuffer;
		}
		else if(pBulkOrInterruptTransfer->TransferBufferMDL)
		{
			BufPtr = (PUCHAR)MmGetSystemAddressForMdl(pBulkOrInterruptTransfer->TransferBufferMDL);
		}
		if (BufPtr==NULL)
		{	// Confused.  Abort.
			return STATUS_SUCCESS;
		}						
		
		
		// Compute data structures.
		// Store system time.
		KeQuerySystemTime(&LocalTime);
									
		
		// Expand hat1.
		switch (BufPtr[0x0a]&0xF0)
		{
			case 0:
				// Do nothing. ;)
				break;
			case 0x10:
				BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x04;
				break;
			case 0x20:
				BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x04+0x08;
				break;
			case 0x30:
				BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x08;
				break;
			case 0x40:
				BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x08+0x10;
				break;
			case 0x50:
				BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x10;
				break;
			case 0x60:
				BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x10+0x20;
				break;
			case 0x70:
				BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x20;
				break;
			case 0x80:
				BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x20+0x04;
				break;
		}		
		
		
		// Compute deltas.
		for(i=0;i<0x0b;i++)
		{
			Delta[i]=DeviceExtension->LastPosition[i] ^ BufPtr[i];
			DeviceExtension->LastPosition[i] = BufPtr[i];
		}
		
		// Load configurations.
		if (BufPtr[8]&0x01)
			switch1 = 1;
		else if (BufPtr[8]&0x02)
			switch1 = 2;
		else
			switch1 = 3;
			
		if (BufPtr[8]&0x08)
			switch2 = 1;
		else if (BufPtr[8]&0x10)
			switch2 = 2;
		else
			switch2 = 3;
			
		shift = (BufPtr[7]&0x40)?2:1;
			
		// Store configuration.
		DeviceExtension->KeyboardStatus[25]=(switch1<<4)+(switch2<<2)+shift;

		// Data structures computed.  Begin engine.

		pos=*(unsigned long*)&BufPtr[7];
		DbgPrn(("Buttons: %X ",pos));
		pos=*(unsigned long*)&Delta[7];
		DbgPrn(("Delta: %X   Mode: %d   Aux: %d    Shift: %d\n",pos,switch1,switch2,shift));
		
		// Script queueing engine.
		// Buttons.
		for(i=0;i<8;i++)
		{
			if ((Delta[7]>>i)&0x01) // Button toggled
				if ((BufPtr[7]>>i)&0x01) // Button is active
				{
					// Button pressed.
					AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Button[i].Pressed,LocalTime);								
				}
				else
				{
					// Button released.
					AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Button[i].Released,LocalTime);
				}
		}

		// Switches.
		// Each 3-way pressed, released, and hold only has one value per configuration.
		// We need to identify the previous configuration with regards to the switch in question so we can perform its release.
		for(i=0;i<2;i++)
		{
			if ((Delta[8]>>(i*3))&0x07) // State has changed.  Will be two deltas.  Deltas ^ current = released, delta & current = pressed.
			{
				unsigned char pressed, released;
				pressed = ((Delta[8]>>(i*3))&0x07) & ((BufPtr[8]>>(i*3))&0x07);
				pressed = ((pressed > 3)||(pressed<1)) ? 3 : pressed;
				released = ((Delta[8]>>(i*3))&0x07) ^ ((BufPtr[8]>>(i*3))&0x07);
				released = ((released > 3)||(released<1)) ? 3 : released;
				DbgPrn(("Switch %d:   Pressed: %d   Released: %d\n",i,pressed,released));
				if (i==0) // Stupid but what can I do?  0 is the mode switch == switch1.
				{
					AddAllScripts(DeviceExtension,pressed,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Slider[i].Pressed,LocalTime);
					AddAllScripts(DeviceExtension,released,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Slider[i].Released,LocalTime);
				}
				else
				{
					AddAllScripts(DeviceExtension,switch1,pressed,shift,&DeviceExtension->Configurations[0][0][0]->Slider[i].Pressed,LocalTime);
					AddAllScripts(DeviceExtension,switch1,released,shift,&DeviceExtension->Configurations[0][0][0]->Slider[i].Released,LocalTime);
				}
			}
		}



		for(i=0;i<4;i++) // Hat switches.
		{
			unsigned char mode;
			unsigned long bitdelta;
			
			// Load mode.  Ignore any but 0 - mode cannot change with config, due to overriding.
			mode = DeviceExtension->Configurations[0][0][0]->Hat[i].Mode;

			// Grab hat bitmap.
			pos=*(unsigned long*)&BufPtr[7];
			b = (pos >> ((i*4)+14))&0x0000000F;
			pos=*(unsigned long*)&Delta[7];
			bitdelta = (pos >> ((i*4)+14))&0x0000000F;
			
			if (mode&0x01) // 8-way
			{
				// reencode for ease of use.
				bitdelta = b ^ bitdelta;
				switch(b)
				{
					case 0x00:
						b = 0; break;
					case 0x01:
						b = 1; break;
					case 0x03:
						b = 2; break;
					case 0x02:
						b = 3; break;
					case 0x06:
						b = 4; break;
					case 0x04:
						b = 5; break;
					case 0x0C:
						b = 6; break;
					case 0x08:
						b = 7; break;
					case 0x09:
						b = 8; break;
				}
				switch(bitdelta)
				{
					case 0x00:
						bitdelta = 0; break;
					case 0x01:
						bitdelta = 1; break;
					case 0x03:
						bitdelta = 2; break;
					case 0x02:
						bitdelta = 3; break;
					case 0x06:
						bitdelta = 4; break;
					case 0x04:
						bitdelta = 5; break;
					case 0x0C:
						bitdelta = 6; break;
					case 0x08:
						bitdelta = 7; break;
					case 0x09:
						bitdelta = 8; break;
				}							
				
				
				if (bitdelta != b)
				{
					AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Hat[i].Pressed[b],LocalTime);
					AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Hat[i].Released[bitdelta],LocalTime);
				}
			}
			else // 4 way
			{
				for (pos=0;pos<4;pos++)
					if ((bitdelta>>pos) & 0x01) // Toggled
						if ((b>>pos) & 0x01) // Pressed
							AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Hat[i].Pressed[pos+1],LocalTime);
						else
							AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Hat[i].Released[pos+1],LocalTime);
			
				if (b == 0)
				{
					if (bitdelta)
						AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Hat[i].Pressed[0],LocalTime);
				}
				else
					if ((b ^ bitdelta)==0)
						AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Hat[i].Released[0],LocalTime);
			}
		}					

		// DbgPrn(("Button Data: %02x %02x %02x %02x\n",BufPtr[7],BufPtr[8],BufPtr[9],BufPtr[10]));

		RtlMoveMemory(&BufPtr[1],BufPtr,pBulkOrInterruptTransfer->TransferBufferLength);

		BufPtr[0]=0x01;
		
		pBulkOrInterruptTransfer->TransferBufferLength+=1;

		
		// Reencode Hat1 status.  Use shift as temp val.
		shift=0;
		pos=*(unsigned long*)&BufPtr[8];
		DbgPrn(("Hat reencoding:    pos = %X\n",pos));
		for(i=0;i<4;i++)
		{
			// Hat mode only valid for 000 config for now.
			if (DeviceExtension->Configurations[0][0][0]->Hat[i].Mode&0x02)
			{
				b = (pos >> ((i*4)+14))&0x0000000F;
				DbgPrn(("Hat reencoding:    Hat %d, b = %X\n",i,b));
				switch(b)
				{
					case 0x00:
						shift = 0; break;
					case 0x01:
						shift = 1; break;
					case 0x03:
						shift = 2; break;
					case 0x02:
						shift = 3; break;
					case 0x06:
						shift = 4; break;
					case 0x04:
						shift = 5; break;
					case 0x0C:
						shift = 6; break;
					case 0x08:
						shift = 7; break;
					case 0x09:
						shift = 8; break;
				}
			}
		}
		DbgPrn(("Hat reencoding:    shift = %d\n",shift));

		RtlCopyMemory(&BufPtr[8],&DeviceExtension->KeyboardStatus[21],4);
		BufPtr[0x0B]=(BufPtr[0x0B]&0x03)+(shift<<4);

		
		//
		// Respawn timer if we have any events.
		//
		QueuePeek(&DeviceExtension->ActionQueue, &Current);
		if (Current.Head)
		{
       		KeSetTimer( &DeviceExtension->Timer, Current.Time, &DeviceExtension->TimerDPC );              
		}
  		
			
			// Fix MDL.
//					if(Temp = (PUCHAR)MmGetSystemAddressForMdl(pBulkOrInterruptTransfer->TransferBufferMDL))
//						RtlCopyMemory(Temp,BufPtr,pBulkOrInterruptTransfer->TransferBufferLength);
	}
	DbgPrn(("JoyPendingActionCount: %d\n",DeviceExtension->pendingActionCount));
	return STATUS_SUCCESS;
}


VOID ExecuteScript( IN PKDPC Dpc, IN PVOID DeferredContext, IN PVOID SystemArgument1, IN PVOID SystemArgument2)
{
struct DEVICE_EXTENSION * DeviceExtension = (struct DEVICE_EXTENSION *)DeferredContext;
    LARGE_INTEGER SysTime;
    QENTRY CurrentQ; PSCRIPT Current;
    unsigned char Offset, Op;
//    SCRIPT Temp, Next;

	// Check for new configuration.   Will not affect existing queue; moved here because of IRQL problems.
//	ReloadProfile(DeviceExtension);


	// Dequeue items.
	KeQuerySystemTime(&SysTime);
	SysTime = RtlLargeIntegerAdd(SysTime,RtlConvertLongToLargeInteger(20)); // Offset because of Windows' timer inaccuracy.
	
	QueuePeek(&DeviceExtension->ActionQueue, &CurrentQ);
	while ((CurrentQ.Head != NULL) && (RtlLargeIntegerGreaterThan(SysTime,CurrentQ.Time)))
	{
		QueuePop(&DeviceExtension->ActionQueue, &CurrentQ);
		Current = CurrentQ.Current;
		
		if (CurrentQ.Control & 0x01) // If we're a hold..
		{
			// somehow compute held status and confirm.
			// Currently just abort.
			Current=NULL;
		}

		
		if (Current != NULL)
		{
			DbgPrn(("Executing item:   Op: %d   Offset: %d   Time: %x %x\n",Current->Op,Current->Offset,CurrentQ.Time.HighPart,CurrentQ.Time.LowPart));		
			Op = Current->Op;
			Offset = Current->Offset;
			
			if (Op & 0x04)  // Control command.
			{
				// Do nothing for now.
				;
			}
			else  // Regular command.
			{
				if (Op & 0x02)
				{
					if (Offset > 32) Offset = 32;
					Offset+=(21*8)-1;
					DeviceExtension->NewData = DeviceExtension->NewData | 1; // New stick data
				}
				else
					DeviceExtension->NewData = DeviceExtension->NewData | 2; // New keybd data
				
				if (Op & 0x01)
				{
					// Release.
					DeviceExtension->KeyboardStatus[(Offset)/8]=DeviceExtension->KeyboardStatus[(Offset)/8] & ~(1 << (Offset%8));
				}
				else
				{
					// Press.
					DeviceExtension->KeyboardStatus[(Offset)/8]=DeviceExtension->KeyboardStatus[(Offset)/8] | (1 << (Offset%8));
				}
			}
			
			if (Current->Next != NULL)
			{
				CurrentQ.Current = Current->Next;
				CurrentQ.Time = RtlLargeIntegerAdd(CurrentQ.Time, RtlConvertLongToLargeInteger(CurrentQ.Current->Delay));
				AddScript(&DeviceExtension->ActionQueue, CurrentQ);
			}
		}
		QueuePeek(&DeviceExtension->ActionQueue, &CurrentQ);
	}
	
	
	// Reset timer.
	if (CurrentQ.Head)
		KeSetTimer( &DeviceExtension->Timer, CurrentQ.Time, &DeviceExtension->TimerDPC );              
		

  	if (DeviceExtension->NewData)
  	{
	  		KeInsertQueueDpc(&DeviceExtension->ReturnDPC,NULL,NULL);
  	}

}



void CompleteRead( IN PKDPC Dpc, IN PVOID DeferredContext, IN PVOID SystemArgument1, IN PVOID SystemArgument2)
{
	struct DEVICE_EXTENSION * DeviceExtension = (struct DEVICE_EXTENSION *)DeferredContext;
	PIRP Irp = NULL;
    PIO_STACK_LOCATION IrpStack;
    ULONG bytesRead;
    UCHAR buffer[30];
    unsigned char switch1,switch2,shift;
    unsigned long b,i,pos;
    

	if (DeviceExtension)
	{
		KIRQL OldIrqL;
				
		KeAcquireSpinLockAtDpcLevel(&DeviceExtension->ReadLock);
		Irp = DeviceExtension->ReadIrp;
		DeviceExtension->ReadIrp = NULL;
		KeReleaseSpinLockFromDpcLevel(&DeviceExtension->ReadLock);
	}

	if (Irp)
	{
		IrpStack = IoGetCurrentIrpStackLocation(Irp);
		
		if ((DeviceExtension->NewData & 0x01) || (Dpc == NULL))
		{	
			DbgPrn(("Completing joystick read..\n"));
			
			buffer[0]=0x01;
			RtlCopyMemory(&buffer[1],DeviceExtension->LastPosition,0x0b);
			RtlCopyMemory(&buffer[8],&DeviceExtension->KeyboardStatus[21],4);
			
			// Reencode Hat1 status.  Use shift as temp val.
			shift=0;
			pos=*(unsigned long*)&buffer[8];
			DbgPrn(("Hat reencoding:    pos = %X\n",pos));
			for(i=0;i<4;i++)
			{
				// Hat mode only valid for 000 config for now.
				if (DeviceExtension->Configurations[0][0][0]->Hat[i].Mode == 0x03)
				{
					b = (pos >> ((i*4)+14))&0x0000000F;
					DbgPrn(("Hat reencoding:    Hat %d, b = %X\n",i,b));
					switch(b)
					{
						case 0x00:
							shift = 0; break;
						case 0x01:
							shift = 1; break;
						case 0x03:
							shift = 2; break;
						case 0x02:
							shift = 3; break;
						case 0x06:
							shift = 4; break;
						case 0x04:
							shift = 5; break;
						case 0x0C:
							shift = 6; break;
						case 0x08:
							shift = 7; break;
						case 0x09:
							shift = 8; break;
					}
				}
			}
			DbgPrn(("Hat reencoding:    shift = %d\n",shift));
	
	
			buffer[0x0B]=(buffer[0x0B]&0x03)+(shift<<4);
			bytesRead=0x0c;
			DeviceExtension->NewData=DeviceExtension->NewData&0xFE;
		}
		else	
		{
			
		    DbgPrn(("Completing keyboard read..\n"));
	 			
			RtlCopyMemory(&buffer[1],DeviceExtension->KeyboardStatus,29);
			buffer[0]=0x02;
		
			RtlFillMemory(&buffer[20],4,0x00);	
	
			// Complete our IO report.
	    	bytesRead = 30;			
	    	DeviceExtension->NewData=DeviceExtension->NewData&0xFD;
		}
		    	
		if (bytesRead > IrpStack->Parameters.DeviceIoControl.OutputBufferLength)
	    {
	       	bytesRead = IrpStack->Parameters.DeviceIoControl.OutputBufferLength;
	    }
	
		RtlCopyMemory((PUCHAR) Irp->UserBuffer, buffer, bytesRead);
	    Irp->IoStatus.Information = bytesRead;
	    Irp->IoStatus.Status = STATUS_SUCCESS;
		
	    IoCompleteRequest(Irp, IO_NO_INCREMENT);
	    DecrementPendingActionCount(DeviceExtension);
	}
}

// Stub function for Iotimer event to send periodic keyboard reports regardless of stick status.
void KeyboardTimer(IN PDEVICE_OBJECT DeviceObject, IN PVOID Context)
{
		CompleteRead( NULL, Context, NULL, NULL);	
}