// QUEUE.C
//
// Handles the timer queue for the X36 scripting engine.
// 
// includes:
//
// queueinit: inits queue.
// scriptadd: adds a whole script to the queue.
// pop: removes the first item from the queue.
// peek: returns the first item in the queue without destroying it
// queueclear: destroys a queue.

// #define DBG 1
#include <WDM.H>
#include "filter.h"


NTSTATUS QueueInit(PXQUEUE MyQueue)
{
	unsigned int i;
	
	if (MyQueue == NULL)
		return STATUS_INVALID_PARAMETER;
	
	MyQueue->Head = NULL;
	KeInitializeSpinLock(&MyQueue->Lock);
	MyQueue->Next = 0; MyQueue->Last = 0;
	for (i=0; i<256; i++)
		MyQueue->Free[i]=(unsigned char)i;
	
	
	return STATUS_SUCCESS;
}

NTSTATUS AddScript(PXQUEUE MyQueue, QENTRY NewEntry)
{
	PQENTRY QNew, QCurrent, QPrev;
	
	
	DbgPrn(("Entering AddScript.\n"));
	if (MyQueue == NULL)
		return STATUS_INVALID_PARAMETER;
	
	if (NewEntry.Current == NULL)
		return STATUS_SUCCESS;
		
	KeAcquireSpinLockAtDpcLevel(&MyQueue->Lock);
	DbgPrn(("Spinlock aquired.\n"));

	if (MyQueue->Next - MyQueue->Last < 255)
	{
		DbgPrn(("Assigning new space in buffer.."));
		QNew = &MyQueue->Buffer[MyQueue->Free[MyQueue->Next]];
		MyQueue->Next++;
		DbgPrn(("..assigned.\n"));
		RtlCopyMemory(QNew,&NewEntry,sizeof(QENTRY));
		QNew->Next = NULL;
		DbgPrn(("Queuing item:   Head: %x   Current: %x   Time: %x %x\n",QNew->Head,QNew->Current,QNew->Time.HighPart,QNew->Time.LowPart));		
		
		if (MyQueue->Head == NULL)
			MyQueue->Head = QNew;
		else
		{
			QCurrent=MyQueue->Head;
			QPrev=NULL;			
			
			while((QCurrent != NULL) && (RtlLargeIntegerGreaterThan(QNew->Time,QCurrent->Time)))
			{
				QPrev = QCurrent;
				QCurrent = QCurrent -> Next;
			}
			
			//DbgPrn(("Queue position:   Head: %x  Prev: %x   New: %x   Current: %x\n",MyQueue->Head,QPrev,QNew,QCurrent));
			if (QPrev != NULL)
			{
				QNew->Next = QCurrent;
				QPrev->Next = QNew;
			}
			else
			{
				MyQueue->Head = QNew;
				QNew->Next = QCurrent;				
			}
			//DbgPrn(("Queue status:   Head: %x   Next: %x\n",MyQueue->Head,(MyQueue->Head==NULL)?NULL:MyQueue->Head->Next));
		}		
	}	
	else
	{
		DbgPrn(("Addscript aborted due to insufficent resources.\n"));
		KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
		DbgPrn(("Spinlock released.\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	
	KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
	DbgPrn(("Spinlock released.\n"));
	return STATUS_SUCCESS;
}



void QueuePop(PXQUEUE MyQueue, PQENTRY Output)
{
	PQENTRY QCurrent, QReturn;
	KIRQL OldIrqL;
	
	if ((MyQueue == NULL) || (Output==NULL))
		return;
		
	KeAcquireSpinLockAtDpcLevel(&MyQueue->Lock);
	
	QCurrent = MyQueue->Head;
	if (QCurrent != NULL)
	{
		MyQueue->Head = QCurrent->Next;
		MyQueue->Free[MyQueue->Last]=(QCurrent - &MyQueue->Buffer[0])/sizeof(QENTRY);
		MyQueue->Last++;
		RtlCopyMemory(Output,QCurrent,sizeof(QENTRY));	
	}
	else
		Output->Head=NULL;
	
	
	KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
}

void QueuePeek(PXQUEUE MyQueue, PQENTRY Output)
{
	if ((MyQueue == NULL) || (Output==NULL))
		return;
		
	KeAcquireSpinLockAtDpcLevel(&MyQueue->Lock);
	
	if (MyQueue->Head != NULL)
		RtlCopyMemory(Output,MyQueue->Head,sizeof(QENTRY));
	else
		Output->Head = NULL;
	
	KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
}

void QueueClear(PXQUEUE MyQueue)
{
	PQENTRY QCurrent;
	KIRQL OldIrqL;	
	
	if (MyQueue == NULL)
		return;
		
	KeAcquireSpinLockAtDpcLevel(&MyQueue->Lock);

	while(MyQueue->Head != NULL)
	{
		QCurrent = MyQueue->Head;
		MyQueue->Head = QCurrent->Next;
		MyQueue->Free[MyQueue->Last]=(QCurrent - &MyQueue->Buffer[0])/sizeof(QENTRY);		
		MyQueue->Last++;
	}
	
	KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
}


void AddAllScripts(struct DEVICE_EXTENSION *devExt, unsigned char mode, unsigned char aux, unsigned char shift, void* Config0Pos, LARGE_INTEGER Time)
{
	unsigned int Delta;
	unsigned char i,j,k;
	QENTRY QNew;
	unsigned char Command;
	
	Delta = (unsigned char*)Config0Pos - (unsigned char*)devExt->Configurations[0][0][0];
//	DbgPrn(("AddAllScripts delta: %d   mode: %d   aux: %d   shift: %d\n",Delta,mode,aux,shift));
//	DbgPrn(("AAS Computed: %X   Actual: %X\n",(((unsigned char*)devExt->Configurations[0][0][0])+Delta),&(devExt->Configurations[0][0][0]->Button[0].Released)));

	for(i=0;i<2;i++)
		for(j=0;j<2;j++)
			for(k=0;k<2;k++)
				if (devExt->Configurations[i*mode][aux*j][shift*k])
				{
					DbgPrn(("AddAllScripts adding:  delta: %d   mode: %d   aux: %d   shift: %d\n",Delta,mode*i,aux*j,shift*k));
					Command = *(((unsigned char*)devExt->Configurations[i*mode][j*aux][k*shift]) + Delta);
					
					if (devExt->Commands[Command] != NULL)
					{
						QNew.Head = QNew.Current = devExt->Commands[Command];
						QNew.Trigger = Delta;
						QNew.Control=((i*mode&0x03)<<6)|((j*aux&0x03)<<4)|((k*shift&0x03)<<2);
						QNew.Time = RtlLargeIntegerAdd(RtlConvertLongToLargeInteger(QNew.Head->Delay),Time);
						AddScript(&devExt->ActionQueue, QNew);
					}
				}
				
	DbgPrn(("AddAllScripts ending.\n"));
}